@@ -143,6 +143,8 @@ struct appfs_pathinfo {
 		} dir;
 		struct {
 			int executable;
+			int suidRoot;
+			int worldaccessible;
 			off_t size;
 		} file;
 		struct {
@@ -812,7 +814,7 @@ static void appfs_get_path_info_cache_flush(uid_t uid, int new_size) {
 static int appfs_get_path_info(const char *path, struct appfs_pathinfo *pathinfo) {
 	Tcl_Interp *interp;
 	Tcl_Obj *attrs_dict, *attr_value;
-	const char *attr_value_str;
+	const char *attr_value_str, *attr_value_str_i;
 	Tcl_WideInt attr_value_wide;
 	int attr_value_int;
 	static __thread Tcl_Obj *attr_key_type = NULL, *attr_key_perms = NULL, *attr_key_size = NULL, *attr_key_time = NULL, *attr_key_source = NULL, *attr_key_childcount = NULL, *attr_key_packaged = NULL;
@@ -934,6 +936,8 @@ static int appfs_get_path_info(const char *path, struct appfs_pathinfo *pathinfo
 				pathinfo->type = APPFS_PATHTYPE_FILE;
 				pathinfo->typeinfo.file.size = 0;
 				pathinfo->typeinfo.file.executable = 0;
+				pathinfo->typeinfo.file.suidRoot = 0;
+				pathinfo->typeinfo.file.worldaccessible = 0;
 
 				Tcl_DictObjGet(interp, attrs_dict, attr_key_size, &attr_value);
 				if (attr_value != NULL) {
@@ -946,8 +950,21 @@ static int appfs_get_path_info(const char *path, struct appfs_pathinfo *pathinfo
 				Tcl_DictObjGet(interp, attrs_dict, attr_key_perms, &attr_value);
 				if (attr_value != NULL) {
 					attr_value_str = Tcl_GetString(attr_value);
-					if (attr_value_str[0] == 'x') {
-						pathinfo->typeinfo.file.executable = 1;
+					for (attr_value_str_i = &attr_value_str[0]; *attr_value_str_i != '\0'; attr_value_str_i++) {
+						switch (*attr_value_str_i) {
+							case 'x':
+								pathinfo->typeinfo.file.executable = 1;
+
+								break;
+							case 'U':
+								pathinfo->typeinfo.file.suidRoot = 1;
+
+								break;
+							case '-':
+								pathinfo->typeinfo.file.worldaccessible = 1;
+
+								break;
+						}
 					}
 				}
 				break;
@@ -1135,6 +1152,7 @@ static int appfs_fuse_readlink(const char *path, char *buf, size_t size) {
 
 static int appfs_fuse_getattr(const char *path, struct stat *stbuf) {
 	struct appfs_pathinfo pathinfo;
+	int changeOwnerToUserIfPackaged;
 	int retval;
 
 	retval = 0;
@@ -1173,20 +1191,33 @@ static int appfs_fuse_getattr(const char *path, struct stat *stbuf) {
 	stbuf->st_ino   = pathinfo.inode;
 	stbuf->st_mode  = 0;
 
+	changeOwnerToUserIfPackaged = 1;
+
 	switch (pathinfo.type) {
 		case APPFS_PATHTYPE_DIRECTORY:
 			stbuf->st_mode = S_IFDIR | 0555;
 			stbuf->st_nlink = 2 + pathinfo.typeinfo.dir.childcount;
 			break;
 		case APPFS_PATHTYPE_FILE:
+			stbuf->st_mode = S_IFREG | 0444;
+
 			if (pathinfo.typeinfo.file.executable) {
-				stbuf->st_mode = S_IFREG | 0555;
-			} else {
-				stbuf->st_mode = S_IFREG | 0444;
+				stbuf->st_mode |= 0111;
+			}
+
+			if (pathinfo.typeinfo.file.suidRoot) {
+				changeOwnerToUserIfPackaged = 0;
+
+				stbuf->st_mode |= 04000;
+			}
+
+			if (pathinfo.typeinfo.file.worldaccessible) {
+				stbuf->st_mode &= ~077;
 			}
 
 			stbuf->st_nlink = 1;
 			stbuf->st_size = pathinfo.typeinfo.file.size;
+
 			break;
 		case APPFS_PATHTYPE_SYMLINK:
 			stbuf->st_mode = S_IFLNK | 0555;
@@ -1213,7 +1244,7 @@ static int appfs_fuse_getattr(const char *path, struct stat *stbuf) {
 			break;
 	}
 
-	if (pathinfo.packaged) {
+	if (pathinfo.packaged && changeOwnerToUserIfPackaged) {
 		stbuf->st_uid   = appfs_get_fsuid();
 		stbuf->st_gid   = appfs_get_fsgid();
 		stbuf->st_mode |= 0200;
@@ -1984,6 +2015,13 @@ static int appfs_opt_parse(int argc, char **argv,  struct fuse_args *args) {
 	if (getuid() == 0) {
 		fuse_opt_parse(args, NULL, NULL, NULL);
 		fuse_opt_add_arg(args, "-oallow_other");
+
+		/*
+		 * This should generally be avoided, but if there are security
+		 * concerns suid can be disabled completely on the commandline
+		 */
+		fuse_opt_parse(args, NULL, NULL, NULL);
+		fuse_opt_add_arg(args, "-osuid");
 	}
 
 	while ((ch = getopt(argc, argv, "dfshvo:")) != -1) {
@@ -2131,6 +2169,11 @@ int main(int argc, char **argv) {
 	argc--;
 	argv++;
 
+	/*
+	 * Set appropriate umask
+	 */
+	umask(022);
+
 	/*
 	 * Set global variables, these should be configuration options.
 	 */
