@@ -40,6 +40,19 @@ namespace eval ::appfs::user {
 	proc get_homedir {} {
 		return [::appfsd::get_homedir]
 	}
+
+	# User-replacable function to update permissions
+	proc change_perms {file sha1 perms} {
+		if {[info exists ::appfs::user::add_perms($file)]} {
+			append perms $::appfs::user::add_perms($file)
+		}
+
+		if {[info exists ::appfs::user::add_perms($sha1)]} {
+			append perms $::appfs::user::add_perms($sha1)
+		}
+
+		return $perms
+	}
 }
 
 namespace eval ::appfs {
@@ -438,7 +451,12 @@ E5AnJIlOnd/tGe0Chf0sFQg+l9nNiNrWGgzdd9ZPJK4=
 					}
 					"file" {
 						set fileInfo(size) [lindex $work 0]
-						set fileInfo(perms) [lindex $work 1]
+
+						# We lower-case the permissions because upper-case permissions
+						# should not be set remotely as they may influence the security
+						# of the system.
+						set fileInfo(perms) [string tolower [lindex $work 1]]
+
 						set fileInfo(sha1) [lindex $work 2]
 
 						set work [lrange $work 3 end]
@@ -750,11 +768,15 @@ E5AnJIlOnd/tGe0Chf0sFQg+l9nNiNrWGgzdd9ZPJK4=
 							"file" {
 								set retval(type) "file"
 								set retval(size) $localpathinfo(size)
+
+								# Once the user writes to a file, all its other
+								# attributes (such as suid) are lost
+
 								_as_user {
 									if {[file executable $localpath]} {
-										set retval(perms) "x"
+										set retval(perms) "x-"
 									} else {
-										set retval(perms) ""
+										set retval(perms) "-"
 									}
 								}
 							}
@@ -787,7 +809,16 @@ E5AnJIlOnd/tGe0Chf0sFQg+l9nNiNrWGgzdd9ZPJK4=
 							array set retval [list type directory]
 						}
 
-						::appfs::db eval {SELECT type, time, source, size, perms FROM files WHERE package_sha1 = $pathinfo(package_sha1) AND file_directory = $directory AND file_name = $file;} retval {}
+						::appfs::db eval {SELECT type, time, source, size, perms, file_sha1 FROM files WHERE package_sha1 = $pathinfo(package_sha1) AND file_directory = $directory AND file_name = $file;} retval {}
+
+						# Allow an administrator to supply additional permissions to remote files
+						if {[info exists retval(perms)]} {
+							# Lower case this in case an upper-cased value was put in
+							# the database before we started lowercasing them
+							set retval(perms) [string tolower $retval(perms)]
+
+							set retval(perms) [::appfs::user::change_perms $path $retval(file_sha1) $retval(perms)]
+						}
 
 						if {[info exists retval(type)] && $retval(type) == "directory"} {
 							set retval(childcount) [llength [getchildren $path]]
